#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    ToXMPP: The Tox<->XMPP Transport/Gateway SleekXMPP-based Component
    
    Copyright (C) 2015  Michał "rysiek" Woźniak <rysiek@hackerspace.pl>
    ToxID: 3FA2E5273F0C368576FE120B374664E3B41E2CDF21639AFED3DC301490FFB01FAAA47B78D5F4
    This file is part of ToXMPP.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from optparse import OptionParser
from toxnode import ToxNode
import multiprocessing as mp
from time import sleep
from glob import iglob
import os

# local debug server when running under Docker
DEBUG_SERVER =  [
                  "172.17.42.1",
                  33445,
                  "3FA2E5273F0C368576FE120B374664E3B41E2CDF21639AFED3DC301490FFB01FAAA47B78D5F4" # change to your ID when testing!
                ]


# class for managing all the Tox instances we need to handle our work
class ToxManager(mp.Process):
    
    # data passing queues to use
    # x2t_queue -> XMPP to Tox queue, passing messages and presence information from XMPP to Tox
    # t2x_queue -> Tox to XMPP queue, passing messages and presence information from Tox to XMPP
    x2t_queue = False
    t2x_queue = False
    
    # mp.Event() flag, if true -- it's time to die.
    teardown = False
    
    # Tox datadir (containing individual Tox instance datafiles, named simply <jid>) to use
    datadir = False
    
    # internal Tox nodes, JID-indexed
    toxnodes = {}
    
    def __init__(self, datadir, x2t_queue, t2x_queue, teardown):
        """
        x2t_queue -> XMPP to Tox queue, passing messages and presence information from XMPP to Tox
        t2x_queue -> Tox to XMPP queue, passing messages and presence information from Tox to XMPP
        """
        
        # parent setup
        super(ToxManager, self).__init__()
        
        # set queues and events
        print('DEBUG :: ToxManager.__init__()')
        self.x2t_queue = x2t_queue
        self.t2x_queue = t2x_queue
        self.teardown = teardown
        
        # set the data directory
        self.datadir = os.path.abspath(datadir)
        
        # create the ToxNodes
        self._create_toxnodes()

  
    # handling delete in a mature way
    def __del__(self):
        # yup, just that for now
        self._teardown()


    # we need some closure
    #
    # to_jid is not available in message data passed from ToxNode,
    # we need to work around that by setting separate, crafted callbacks
    # per each ToxNode instance
    def _generate_toxnode_msgcallback(self, to_jid):
        # that's our callback
        def cb(from_toxid, toxmsg):
            # and that's the actual handler
            return self.handle_tox_message(from_toxid, to_jid, toxmsg)
        # closure acquired!
        return cb
      
    
    # creating toxnodes based on datafiles found in datadir
    def _create_toxnodes(self):
      
        # get all the data files and instantiate all ToxNodes needed
        for datafile in iglob(os.path.join(self.datadir, '*')):
          
            # what's the JID? hint, it's the filename
            jid =  os.path.basename(datafile)
            print('DEBUG :: +-- creating a ToxNode for %s' % jid)
            
            #self.toxnodes[jid] = ToxNode(datafile=datafile, servers=[DEBUG_SERVER], message_received_callback=self._generate_toxnode_msgcallback(jid))
            self.toxnodes[jid] = ToxNode(datafile=datafile, message_received_callback=self._generate_toxnode_msgcallback(jid))
            
            print('DEBUG :: +-- done.')


    # run the show
    def run(self):
      
        print('DEBUG :: ToxManager.run()')
        
        try:
            while not self.teardown.is_set():
                # handle queue items
                # TODO: handle too many messages situation
                while not self.x2t_queue.empty():
                    self.handle_x2t_queue_item(self.x2t_queue.get())
                    
                # run main Tox routine for each internal Tox instance
                for jid in self.toxnodes:
                    self.toxnodes[jid].do()
                
                # let's not over.do() it ;)
                sleep(0.03)

            # ah, got the teardown flag -- just inform, proper handling done in finally block below
            else:
                print('DEBUG :: got teardown, tearing down ToxManager')
        
        # we don't want no KeyboardInterrupts here, go away.
        except KeyboardInterrupt:
            print('DEBUG :: ignoring a keyboard interrupt')
            
        # for all other exceptions we need some cleanup/handling code
        finally:
            # inform
            print('DEBUG :: gracefully shutting down all threads')
            # set the teardown flag for all processes
            self._teardown()
            
  
    def _teardown(self):
        # inform
        print('DEBUG :: tearing down all ToxNodes')
        for jid in set(self.toxnodes.keys()):
            print('DEBUG :: +-- %s' % jid)
            del self.toxnodes[jid]
        # set the flag just in case
        print('DEBUG :: making sure the teardown flag is set')
        self.teardown.set()
  
    
    def get_toxnode(self, jid):
        """
        getting an internal toxnode associated with a given (bare) JID
        yes, this will get much more sophisticated
        """
        return self.toxnodes[jid.bare]
    
    
    # setting our own status based on status received from external JID
    def set_own_status(self, status):
        """
        http://aitjcize.github.io/PyTox/pytox.Tox.html#pytox.Tox.set_user_status
        Tox.USERSTATUS_NONE the user is online
        Tox.USERSTATUS_AWAY the user is away
        Tox.USERSTATUS_BUSY the user is busy
        Tox.USERSTATUS_INVALID  invalid status -> I don't think we should ever set it, don't you? ;)
        """
        
        
    def handle_message(self, msg):
        """
        handling the message stanza
        this *should* be a message from an external JID
        to an <external-ToxId>@<own-server-addr>
        TODO: checking if the ToxID is an internal one, by chance
        TODO: checking if the local domain is correct
        """
        
        # get the related Tox node
        tn = self.get_toxnode(msg["from"])
        
        # forward the message
        tn.send_message(
            # we need a friend_id from the ToxNode,
            # and for that we definitely need the user part of the "to" address
            # which is assumed to be a valid ToxID
            tn.get_friend_id(msg["to"].user),
            "%(from)s sent you a message: %(body)s" % msg
          )
    
    
    def handle_subscription(self, qitem):
        """
        handling subsciption
        in response to: 'subscribe', 'subscribed', 'unsubscribe', 'unsubscribed'
        """
    
    
    def handle_probe(self, qitem):
        """
        handling a presence probe
        """
    
    
    def handle_tox_message(self, from_toxid, to_jid, toxmsg):
        # TODO: stubity-stub, put the message data into the t2x_queue
        # from is just a "short" Tox id, the public key *without* nospam and checksum
        # https://github.com/irungentoo/toxcore/issues/854
        # this needs to get handled on the XMPP side, matched against the user's roster
        msg = {
            'message': {
              'to'  : to_jid,
              'from': from_toxid,
              'body': 'Got message from %s:\n%s' % (from_toxid, toxmsg)
            }
          }
        # put it in the Tox to XMPP queue
        self.t2x_queue.put(msg)
    
      
    # handling a queue item
    def handle_x2t_queue_item(self, qitem):
        """
        Possible stanza types:
        presence  - presence info about a particular user, and (un)subscribe notifications
                    https://github.com/fritzy/SleekXMPP/wiki/Stanzas:-Presence
        message   - text message stanza
                    https://github.com/fritzy/SleekXMPP/wiki/Stanzas:-Message
        """
        
        print('DEBUG :: handle_x2t_queue_item()')
        print('DEBUG :: +-- qitem: %s' % str(qitem).encode('utf-8'))
        
        if qitem['message']:
            self.handle_message(qitem['message'])
            
        elif qitem['presence']:
            # TODO : current implementation does not make any sense
            # handling the presence stanza
            # presence status
            # 'xa'  -> "eXtended Away
            # 'dnd' -> "Do Not Disturbed"
            presence = qitem['presence']
            print('DEBUG :: +-- type : %(type)s' % presence)
            if presence['type'] in ('chat', 'available', 'away', 'dnd', 'xa', 'unavailable'):
                self.set_own_status(presence)

            # subscription-related
            elif presence['type'] in ('subscribe', 'subscribed', 'unsubscribe', 'unsubscribed'):
                self.handle_subscription(presence)
                
            # presence probe
            elif presence['type'] == 'probe':
                self.handle_probe(presence)
                
            elif presence['type'] == 'error':
                # we should handle errors... somehow
                pass
        
        # dafuq?
        else:
            # can't work with that!
            raise TypeError('Internal "message" or "presence" datatypes expected!')
        


# just making sure we're tha real deal
if __name__ == '__main__':

    # Setup the command line arguments.
    optp = OptionParser()
    optp.add_option("-d", "--datadir", dest="datadir",
                    help="Datadir to glob for data files")

    opts, args = optp.parse_args()

    if opts.datadir is None:
        opts.datadir = input("Datadir path: ")
        
    # message passing queues
    x2tq = mp.Queue()
    t2xq = mp.Queue()

    # run the Tox process
    tp = ToxManager(opts.datadir, x2tq, t2xq)
    tp.start()

    txt = input('Testing keyboard, lol... Enter anything: ')
    print("Input was:\n%s" % txt)
    
    msg = {
        'message': {
            'to'  : DEBUG_SERVER[2],
            'from':'rysiek@tox-prosody',
            'body': txt
          }
        }
    print("msg is : %s" % msg)
    x2tq.put(msg)
