#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    ToXMPP: The Tox<->XMPP Transport/Gateway SleekXMPP-based Component
    
    Copyright (C) 2015  Michał "rysiek" Woźniak <rysiek@hackerspace.pl>
    ToxID: 3FA2E5273F0C368576FE120B374664E3B41E2CDF21639AFED3DC301490FFB01FAAA47B78D5F4
    This file is part of ToXMPP.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys
import logging
import getpass
from optparse import OptionParser

import sleekxmpp
from sleekxmpp.componentxmpp import ComponentXMPP

import yaml_roster

import multiprocessing as mp
from mptox import ToxManager, DEBUG_SERVER
from time import sleep

# Python versions before 3.0 do not use UTF-8 encoding
# by default. To ensure that Unicode is handled properly
# throughout SleekXMPP, we will set the default encoding
# ourselves to UTF-8.
if sys.version_info < (3, 0):
    from sleekxmpp.util.misc_ops import setdefaultencoding
    setdefaultencoding('utf8')
else:
    raw_input = input


def tdbg(txt):
  print ("TOXMPP :: %s" % txt.encode('utf-8'))        


class ToXMPPComponent(ComponentXMPP):
    """
    Tox <-> XMPP Transport/Gateway compnent
    """
    
    # data passing queues to use
    # x2t_queue -> XMPP to Tox queue, passing messages and presence information from XMPP to Tox
    # t2x_queue -> Tox to XMPP queue, passing messages and presence information from Tox to XMPP
    x2t_queue = False
    t2x_queue = False
    
    # ToxNodes manager
    toxmanager = None
    
    # an mp.Event() flag, used to signal to child processes that we're being killed
    teardown = False
    

    def __init__(self, jid, secret, server, port):
        # parent setup
        ComponentXMPP.__init__(self, jid, secret, server, port)
        # debug
        tdbg('roster\n%s\n' % self.roster);

        # You don't need a session_start handler, but that is
        # where you would broadcast initial presence.

        # The message event is triggered whenever a message
        # stanza is received. Be aware that that includes
        # MUC messages and error messages.
        self.add_event_handler("message", self.message)
        
        #self.add_event_handler("presence_subscribe", self.presence_subscribe)
        # auto-managing subscriptions, for the time being
        self.auto_authorize = None
        self.auto_subscribe = None
        tdbg('setting event handlers: presence_probe, got_online, got_offline, subscribe, subscribed')
        self.add_event_handler('presence_probe', self.handle_presence_probe)
        self.add_event_handler('got_online', self.handle_got_online)
        self.add_event_handler('got_offline', self.handle_got_offline)
        self.add_event_handler('presence_subscribe', self.handle_presence_subscribe)
        self.add_event_handler('presence_subscribed', self.handle_presence_subscribed)
        self.add_event_handler('presence_unsubscribe', self.handle_presence_unsubscribe)
        self.add_event_handler('presence_unsubscribed', self.handle_presence_unsubscribed)
        # log bound jid
        tdbg('+-- bound jid is: %s' % self.boundjid.bare)
        
        # queue
        self.x2t_queue = mp.Queue()
        self.t2x_queue = mp.Queue()
        
        # teardown Event
        self.teardown = mp.Event()
        self.teardown.clear()
        
        # ToxManager
        self.toxmanager = ToxManager('toxdata/', self.x2t_queue, self.t2x_queue, self.teardown)
        self.toxmanager.start()


    def run(self):
        """
        Main queue processing loop
        processing as long as teardown mp.Event() flag is not set
        """
        try:
            while not self.teardown.is_set():
                
                # handle the Tox to XMPP queue items
                while not self.t2x_queue.empty():
                    self.handle_t2x_queue_item(self.t2x_queue.get())
                    
                # just you wait
                sleep(0.5)
            
            # ah, got the teardown flag -- just inform, proper handling done in finally block below
            else:
                tdbg('got teardown')
            
        # ach, an exception
        finally:
            tdbg('tearing down ToXMPPComponent')
            # yeah, _teardown() contains everything we need to do
            self._teardown()
        
    
    # getting an internal JID from a ToxID (either "short" or a "full" one)
    # for each "short" Tox IDs we might have several different "long" versions
    # so we need to find the one that has a given friend_jid in its roster
    # TODO: this has to get fixed on some other level!
    def get_jid_from_toxid(self, toxid, friend_jid):
      
        # well, if it's a valid "long" Tox id, just use this
        if len(toxid) == 76:
            # if not valid, complain
            if not self._verify_toxid(toxid):
                raise ValueError("Valid (long) Tox ID expected, '%s' received" % toxid)
            # assemble the internal jid
            return '%s@%s' % (toxid, self.boundjid.bare)
          
        # ah, so is it a short tox id?
        if len(toxid) == 64:
            # while we're at it... let's check if it's a valid short tox id
            try:
                int(toxid, 16)
            except ValueError:
                raise ValueError("Valid (short) Tox ID expected, '%s' received" % toxid)
            # doesn't contain 0x in front?
            if toxid.lower()[0:2] == '0x':
                raise ValueError("Valid (short) Tox ID expected, '%s' received" % toxid)
              
            # okay, we have a *valid* short Tox ID on our hands
            # let's loop through internal JIDs and see if any fit
            ijids = []
            for ijid in self.roster:
                # found a matching one?
                # is the external JID in its roster?
                if ijid.startswith(toxid) and friend_jid in self.roster[ijid]:
                    ijids.append(ijid)
            
            # this is where it gets tricky
            # 1. a given external JID can be in several internal JIDs rosters
            # even in several that match a given short Tox ID
            # this could happen if an external JID user added a given Tox user
            # to their roster several times with the Tox user changing their
            # nospam in the meantime
            if ijids:
                # well, at least we've found *something*; is there more than one matching jid?
                if len(ijids) == 1:
                    # nope, just return this
                    return ijids[0]
                  
                # more than one? uh-huh. find the first one that has friend_jid in roster
                # TODO: this is going away as soon as the whole short/long ToxID thing gets fixed somewhere else
                for ijid in ijids:
                    if friend_jid in self.roster[ijid]:
                        return ijid
                # well, just return the first one
                else:
                    return ijids[0]
                    
            # 2. the external JID can even be unavailable in any internal JID's
            # rosters
            else:
                return '%s@%s' % (toxid, self.boundjid.bare)
            
        
        # dafuq?
        else:
            raise ValueError("Valid (long or short) Tox ID expected, '%s' received" % toxid)
    
    
    def handle_t2x_queue_item(self, msg):
        """
        handling the message stanza
        this *should* be a message from an ixternal ToxID-based JID
        to an external JID
        """
        
        # yay
        tdbg('toxmpp.run() : we got message on the queue!')
        
        # message? just send it already
        if msg['message']:
            tdbg('toxmpp.run() : it\'s a message!')
            msg = msg['message']
            # we have to handle the "from" part -- it contains only the "short" Tox ID:
            # https://github.com/irungentoo/toxcore/issues/854
            # ...insufficient to build the correct internal JID (these use the "long" version)
            # so we have to match it against rosters containing the 'to' external JID
            msg['from'] = self.get_jid_from_toxid(msg['from'], msg['to'])
            
            # debug info
            tdbg('               +-- from : %(from)s\nTOXMPP ::                +-- to   : %(to)s\nTOXMPP ::                +-- msg  : %(body)s' % msg )
            tdbg('toxmpp.run() : sending the message out');
            self.send_message(mto=msg['to'],
                              mfrom=msg['from'],
                              mbody=msg['body'],
                              mtype='chat') # default type
        # wat
        else:
            raise TypeError('Internal "message" data structure expected expected')


    # verify the toxid
    def _verify_toxid(self, toxid):
        """
        Checking if a given string is a valid ToxID, i.e. is 76 chars
        long hex (i.e. A-Fa-f0-9) string
        https://libtoxcore.so/core_concepts.html
        checking *only* for the long, nospam-containing version!
        cf. https://github.com/irungentoo/toxcore/issues/854
        TODO: verifying the checksum?
        """
        try:
            # hex?
            int(toxid, 16)
            # 76 chars?
            if len(toxid) != 76:
                return False;
            # doesn't contain 0x in front?
            if toxid.lower()[0:2] == '0x':
                  return False;
            # AOK
            return True
        
        except ValueError:
            # apparently, not hex
            return False

    
    def _verify_jid(self, jid):
        """
        Checking if a given (local) JID is correct, i.e. is either
        self.boundjid.bare, or <valid-ToxID>@<self.boundjid.bare>.
        """
        
        # this component's address?
        if jid == self.boundjid.bare:
            return True
          
        # nah, <something>@<something>; let's make sure it's
        # <valid-ToxID>@<self.boundjid.bare>!
        if not self._verify_toxid(jid.user):
            return False
        if jid.domain != self.boundjid.bare:
            return False
        
        # we're done here.
        return True


    def message(self, msg):
        """
        Process incoming message stanzas. Be aware that this also
        includes MUC messages and error messages. It is usually
        a good idea to check the messages's type before processing
        or sending replies.

        Since a component may send messages from any number of JIDs,
        it is best to always include a from JID.

        Arguments:
            msg -- The received message stanza. See the documentation
                   for stanza objects and the Message stanza to see
                   how it may be used.
        """
        
        tdbg('Message from %(from)s to %(to)s' % msg)
        # should we forward the message to Tox?
        if self._verify_toxid(msg['to'].user):
            tdbg('+-- valid ToxID: %s' % msg['to'].user)
            
            # create a new message, in the internal datastructure this time
            toxmsg = {
                'message': {
                    'type'   : msg['type'],
                    'to'     : msg['to'],
                    'from'   : msg['from'],
                    'body'   : msg['body'],
                    'id'     : msg['id'],
                    'subject': msg['subject']
                  }
                }
            # push the message to Tox
            self.x2t_queue.put(toxmsg)
        else:
            tdbg('+-- invalid ToxID: %s' % msg['to'].user)
        
        # The reply method will use the message's 'to' JID as the
        # outgoing reply's 'from' JID.
        msg.reply("Thanks for sending\n%(body)s" % msg).send()


    # handle a contact getting online
    def handle_got_online(self, presence):
        tdbg('got online: %s for %s' % (presence['from'], presence['to']))
        roster_item = self.roster[presence['to']][presence['from']]
        
        # debug
        tdbg('+-- subscription: %s' % roster_item['subscription'])
        tdbg('roster\n%s\n' % self.roster);

        
    # handle a contact getting offline
    def handle_got_offline(self, presence):
        tdbg('got offline: %s for %s' % (presence['from'], presence['to']))
        #roster_item = self.roster[presence['to']][presence['from']]
        # debug
        tdbg('roster\n%s\n' % self.roster);

    # handle a presence probe
    def handle_presence_probe(self, presence):
        tdbg('a presence probe from %s to %s; NOT replying...' % (presence['from'], presence['to']))
        # debug
        tdbg('roster\n%s\n' % self.roster);
    
    
    # handle subscription notification
    def handle_presence_subscribe(self, presence):
        tdbg('a subscription request from %s for %s' % (presence['from'], presence['to']))
        roster_item = self.roster[presence['to']][presence['from']]
        
        # is the "to" JID sane?
        # it should either be a <valid ToxID>@<self.boundjid.bare>
        if not self._verify_jid(presence['to']):
            tdbg('+-- jid fail! unauthorizing...')
            roster_item.unauthorize()
            return False;
        
        # If the subscription request is accepted.
        tdbg('+-- accepting...')
        roster_item.authorize()

        # If a bi-directional subscription does not exist, create one.
        if roster_item['subscription'] not in ('to', 'both'):
            tdbg('+-- subscribing...')
            roster_item.subscribe()
          
        # debug
        tdbg('roster\n%s\n' % self.roster);


    def handle_presence_subscribed(self, presence):
        tdbg('a subscription notification from %s for %s' % (presence['from'], presence['to']))
        roster_item = self.roster[presence['to']][presence['from']]
        
        tdbg('+-- subscription: %s' % roster_item['subscription'])
        # we want to be subscribed to all JIDs that allow us to
        # as long as they're subscribing to sane JIDs on our side
        # so, reality check
        if not self._verify_jid(presence['to']):
            tdbg('+-- jid fail! unauthorizing...')
            roster_item.unauthorize()
            return False;

        # we're here, "our" JID is sane, let's roll
        if roster_item['subscription'] not in ('to', 'both'):
            roster_item.subscribe()
            tdbg('+-- subscribing to: %s' % roster_item['subscription'])
          
        # pretty simple, we accept all and show our status to all
        roster_item.send_presence()
        
        # debug
        tdbg('roster\n%s\n' % self.roster);
    
    #
    def handle_presence_unsubscribe(self, presence):
        # debug
        tdbg('an unsubscribe request from %s for %s' % (presence['from'], presence['to']))
        roster_item = self.roster[presence['to']][presence['from']]
        
        tdbg('+-- subscription: %s' % roster_item['subscription'])
        
        #if not self.backend.subbed_to(presence['from']):
        #   self.sendPresence(pto=presence['from'],
                            #pfrom=presence['to'],
                            #ptype='unsubscribe')
        
        # debug
        tdbg('roster\n%s\n' % self.roster);
    
    # 
    def handle_presence_unsubscribed(self, presence):
        tdbg('an unsubscribed notification from %s for %s' % (presence['from'], presence['to']))
        roster_item = self.roster[presence['to']][presence['from']]
        
        # tit for tat
        if roster_item['subscription'] in ('to', 'both'):
          roster_item.unsubscribe()
        if roster_item['subscription'] in ('from', 'both'):
          roster_item.unauthorize()
        tdbg('+-- subscription: %s' % roster_item['subscription'])
        # Remove the subscription state from the backend
        #tdbg('+-- saving info...')
        #self.backend.unsub_from(presence['from'])
        
        # debug
        tdbg('roster\n%s\n' % self.roster)
    
    
    def _teardown(self):
        # tear down the XMPP comms
        self.disconnect(wait=True)
        # signal the children
        self.teardown.set()
    
        
    # teardown/graceful shutdown
    def __del__(self):
        # debug
        tdbg('tearing down...')
        # start teardown
        self._teardown()
        # join the toxmanager process and give it time to gracefully die
        self.toxmanager.join()


if __name__ == '__main__':
    # Setup the command line arguments.
    optp = OptionParser()

    # Output verbosity options.
    optp.add_option('-q', '--quiet', help='set logging to ERROR',
                    action='store_const', dest='loglevel',
                    const=logging.ERROR, default=logging.INFO)
    optp.add_option('-d', '--debug', help='set logging to DEBUG',
                    action='store_const', dest='loglevel',
                    const=logging.DEBUG, default=logging.INFO)
    optp.add_option('-v', '--verbose', help='set logging to COMM',
                    action='store_const', dest='loglevel',
                    const=5, default=logging.INFO)

    # JID and password options.
    optp.add_option("-j", "--jid", dest="jid",
                    help="JID to use")
    optp.add_option("-p", "--password", dest="password",
                    help="password to use")
    optp.add_option("-s", "--server", dest="server",
                    help="server to connect to")
    optp.add_option("-P", "--port", dest="port",
                    help="port to connect to")
    
    # Tox datadir
    #optp.add_option("-t", "--tox-datadir", dest="tox_datadir",
    #                help="Data directory for Tox")

    opts, args = optp.parse_args()

    if opts.jid is None:
        opts.jid = raw_input("Component JID: ")
    if opts.password is None:
        opts.password = getpass.getpass("Password: ")
    if opts.server is None:
        opts.server = raw_input("Server: ")
    if opts.port is None:
        opts.port = int(raw_input("Port: "))

    #if opts.tox_datadir is None:
    #    opts.tox_datadir = input("Tox datadir path: ")

    # Setup logging.
    logging.basicConfig(level=opts.loglevel,
                        format='%(levelname)-8s %(message)s')

    # Setup the ToXMPPComponent and register plugins. Note that while plugins
    # may have interdependencies, the order in which you register them does
    # not matter.
    xmpp = ToXMPPComponent(opts.jid, opts.password, opts.server, opts.port)
    xmpp.register_plugin('YAMLRoster', module=yaml_roster) # YAML Roster flat-file backend
    xmpp.register_plugin('xep_0030') # Service Discovery
    xmpp.register_plugin('xep_0004') # Data Forms
    xmpp.register_plugin('xep_0060') # PubSub
    xmpp.register_plugin('xep_0199') # XMPP Ping

    # Connect to the XMPP server and start processing XMPP stanzas.
    if xmpp.connect():
        try:
            # start processing in background
            xmpp.process(block=False)
            # start the queue processing loop
            xmpp.run()
        # catch the Ctrl+C
        except KeyboardInterrupt:
            # cleanup
            tdbg('KeyboardInterrupt caught, exiting...')
            del xmpp
            sys.exit(0)
    else:
        print("Unable to connect.")
