# -*- coding: utf-8 -*-
"""
    ToXMPP: The Tox<->XMPP Transport/Gateway SleekXMPP-based Component
    
    Copyright (C) 2015  Michał "rysiek" Woźniak <rysiek@hackerspace.pl>
    ToxID: 3FA2E5273F0C368576FE120B374664E3B41E2CDF21639AFED3DC301490FFB01FAAA47B78D5F4
    This file is part of ToXMPP.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import yaml
import os
from glob import iglob

import sleekxmpp
from sleekxmpp.plugins.base import BasePlugin

def yrdbg(txt):
  print ("YAMLRoster :: %s" % txt)    

class YAMLRoster(BasePlugin):
    """
    Simple roster backend keeping the roster in flat YAML files,
    one file per identity roster / jid combination
    
    Usage:
    import yaml_roster
    xmpp = ClientXMPP() /OR/ xmpp = ComponentXMPP() 
    xmpp.register_plugin('YAMLRoster', module=yaml_roster) # YAML Roster backend
    """
    
    # default config
    default_config = {
      'storage_dir' : './roster/'
    }
    
    # plugin initialization
    def plugin_init(self):
        # debug
        yrdbg('plugin init')
        # name and description
        self.description = 'yaml_roster'
        self.description = 'YAML Roster backend, storing roster(s) in flat YAML files in a given directory'
        # storage directory
        self.storage_dir = os.path.abspath(self.config.get('storage_dir', './roster/'))
        yrdbg('+-- storage dir: %s' % self.storage_dir)
        
        # make sure the directory exists
        if not os.path.exists(self.storage_dir):
            # we need to create it? eh...
            # if it fails, OSError will be raised
            yrdbg('+-- storage dir doesn\'t exist, creating...')
            os.makedirs(self.storage_dir, 0o750)
            yrdbg('+-- created.')
            
        # oh, so it does exist? is it a directory?
        elif not os.path.isdir(self.storage_dir):
            yrdbg('+-- storage dir not a dir')
            raise IOError('Storage path \'%s\' expected to be a directory.' % self.storage_dir)
            
        # do we have the permissions?
        elif not os.access(self.storage_dir, os.R_OK | os.W_OK | os.X_OK):
            yrdbg('+-- storage dir perms wrong')
            raise IOError('Storage directory \'%s\' is not accessible (needed rwx).' % self.storage_dir)
        
        # make it so!
        yrdbg('+-- setting the backend')
        self.xmpp.roster.set_backend(self)

        
    # simple helper function to get a filename
    # based on owner_jid and jid
    def _get_item_path(self, owner_jid, jid):
        # sort out the filename
        # it's safe to use ':' as a delimeter here, because reasons:
        # http://www.xmpp.org/extensions/xep-0106.html#intro
        return os.path.join(self.storage_dir, '%s.yml' % ':'.join((owner_jid, jid)))

    
    # getting the entries' jids
    def entries(self, owner_jid, db_state=None):
        
        # debug
        yrdbg('entries')
        # roster init
        roster = []

        # no owner_jid? just owners' jids, then!
        if owner_jid is None:
            # yup, get all
            yrdbg('+-- getting owners\' jids')
            for fname in os.listdir(self.storage_dir):
                # no need to even read contents of those files,
                # we only need to split the names and get owners' jids from them
                fname = fname.split(':')[0]
                if fname not in roster:
                    yrdbg('+-- found: %s' % fname)
                    roster.append(fname)
                    
        # nah, just owned by owner_jid
        else:
            # some globbing needed
            yrdbg('+-- getting for owner_jid: %s' % owner_jid)
            glob_expr = self._get_item_path(owner_jid, '*')
            yrdbg('+-- glob_expr: %s' % glob_expr)
            for fname in iglob(glob_expr):
                # no need to even read contents of those files,
                # we only need to split the names and get roster jids from them
                # without the '.yml' suffix, of course!
                yrdbg('+-- considering : %s' % fname)
                fname = os.path.basename(fname)
                yrdbg('+-- basename    : %s' % fname)
                fname = fname.split(':')[1]
                yrdbg('+-- split       : %s' % fname)
                fname = fname[:-4]
                yrdbg('+-- ext-less    : %s' % fname)
                if fname not in roster:
                    yrdbg('+-- found: %s' % fname)
                    roster.append(fname)
            
        # return the roster
        return roster

    
    # loading the roster from the datastore
    def load(self, owner_jid, jid, db_state=None):
        # debug
        yrdbg('load :: %s : %s' % (owner_jid, jid))
        # get the filename
        item_file = self._get_item_path(owner_jid, jid)
        yrdbg('+-- item_file: %s' % item_file)
        # get the yaml from file
        try:
          with open(item_file, 'r') as f:
              # yup, it's that simple
              # using safe_load, because we're just getting a dictionary,
              # no need for fancy objects thankyouverymuch!
              yrdbg('+-- file open, loading...')
              return yaml.safe_load(f)
        # if the file is not there, assume empty roster
        except FileNotFoundError:
          return {}

    
    # saving the roster to the datastore
    def save(self, owner_jid, jid, item, db_state=None):
        yrdbg('save :: %s : %s' % (owner_jid, jid))
        # get the filename
        item_file = self._get_item_path(owner_jid, jid)
        yrdbg('+-- item_file: %s' % item_file)
        # save the data to the file
        with open(item_file, 'w') as f:
            # get the yaml and dump it to the file directly
            # oh, and format it nicely, block-ily, while you're at it!
            yrdbg('+-- file open, saving...')
            yaml.dump(item, f, default_flow_style=False)